#ifndef __MATVEC_Fast_CSC_CSR_KERNELS_H__
#define __MATVEC_Fast_CSC_CSR_KERNELS_H__

#define __MY_RESTRICT__ __restrict__

#ifdef __TEST_CPU__
inline void
fproduct
(
  idx_k        const                      nRows, 
  idx_k&                                  iFace,
  idx_k        const                      iRow,
  LDUoffDiag_k const * const __restrict__ upper,
  LDUoffDiag_k const * const __restrict__ lower,
  idx_k        const * const __restrict__ neigh,
  fld_k        const * const __restrict__ psi,
  fld_k              * const __restrict__ Apsi
)
{
  const fld_k& xr = psi[iRow];
  fld_k tmp = 0.0;
  for (idx_k i=0; i<nRows; ++i)
  {
    tmp += upper[iFace]*psi[neigh[iFace]];
    Apsi[neigh[iFace]] += lower[iFace]*xr;
    ++iFace;
  }
  Apsi[iRow] += tmp;

  return;
};

template<const unsigned int N>
inline void
fproduct
(
  idx_k&                                  iFace,
  idx_k        const                      iRow,
  LDUoffDiag_k const * const __restrict__ upper,
  LDUoffDiag_k const * const __restrict__ lower,
  idx_k        const * const __restrict__ neigh,
  fld_k        const * const __restrict__ psi,
  fld_k              * const __restrict__ Apsi
)
{
  const fld_k& xr = psi[iRow];
  fld_k tmp = 0.0;
  #pragma unroll (N)
  for (idx_k i=0; i<N; ++i)
  {
    tmp += upper[iFace]*psi[neigh[iFace]];
    Apsi[neigh[iFace]] += lower[iFace]*xr;
    ++iFace;
  }
  Apsi[iRow] += tmp;
  
  return;
};


void
matVec_Fast_CSC_CSR_CPU
(
  idx_k nRows, // = m_IO.matSize();
  unsigned char const * const  __restrict__ nnz,
  idx_k         const * const  __restrict__ n, //=Permutation.pNeigh();
  LDUoffDiag_k  const * const  __restrict__ u, //=Permutation.upper( randomMatrix.upper() );
  LDUoffDiag_k  const * const  __restrict__ l, //=Permutation.lower( randomMatrix.lower() );
  fld_k         const * const  __restrict__ x, //=Permutation.x(     randomMatrix.x()     );
  fld_k               * const  __restrict__ b  //=m_result;
)
{
  // Local variables
  idx_k iFace = 0;

  // Off diagonal part of the product
  for (int iRow=0; iRow<nRows; ++iRow)
  {
    switch ( nnz[iRow] )
    {
      case (0):       continue;
      case (1):       fproduct<1 >(iFace, iRow, u, l, n, x, b); break;
      case (2):       fproduct<2 >(iFace, iRow, u, l, n, x, b); break;
      case (3):       fproduct<3 >(iFace, iRow, u, l, n, x, b); break;
      case (4):       fproduct<4 >(iFace, iRow, u, l, n, x, b); break;
      case (5):       fproduct<5 >(iFace, iRow, u, l, n, x, b); break;
      case (6):       fproduct<6 >(iFace, iRow, u, l, n, x, b); break;
      case (7):       fproduct<7 >(iFace, iRow, u, l, n, x, b); break;
      case (8):       fproduct<8 >(iFace, iRow, u, l, n, x, b); break;
      case (9):       fproduct<9 >(iFace, iRow, u, l, n, x, b); break;
      case (10):      fproduct<10>(iFace, iRow, u, l, n, x, b); break;
      case (11):      fproduct<11>(iFace, iRow, u, l, n, x, b); break;
      case (12):      fproduct<12>(iFace, iRow, u, l, n, x, b); break;
      case (13):      fproduct<13>(iFace, iRow, u, l, n, x, b); break;
      case (14):      fproduct<14>(iFace, iRow, u, l, n, x, b); break;
      case (15):      fproduct<15>(iFace, iRow, u, l, n, x, b); break;
      case (16):      fproduct<16>(iFace, iRow, u, l, n, x, b); break;
      case (17):      fproduct<17>(iFace, iRow, u, l, n, x, b); break;
      case (18):      fproduct<18>(iFace, iRow, u, l, n, x, b); break;
      case (19):      fproduct<19>(iFace, iRow, u, l, n, x, b); break;
      case (20):      fproduct<20>(iFace, iRow, u, l, n, x, b); break;
      case (21):      fproduct<21>(iFace, iRow, u, l, n, x, b); break;
      default: fproduct(nnz[iRow], iFace, iRow, u, l, n, x, b); break;
    }
  }

  return;
};
#endif

#ifdef __TEST_NATIVE_CUDA__
__global__
void
matVec_Fast_CSC_CSR_nativeCUDA
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif


#ifdef __TEST_OPEN_ACC__
void
matVec_Fast_CSC_CSR_openACC
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#ifdef __TEST_OPEN_MP__
void
matVec_Fast_CSC_CSR_openMP
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#endif
