#ifndef __MATVEC_TCO_CSC_CSR_KERNELS_H__
#define __MATVEC_TCO_CSC_CSR_KERNELS_H__

#define __MY_RESTRICT__ __restrict__
#define NCASES 32

#ifdef __TEST_CPU__

template<const unsigned int N>
inline void
tco_product
(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k                                    iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
);

inline void
tco_product_zero
(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k                                    iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
);

inline void
tco_product_final
(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k                                    iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
);

void (*tape[NCASES])(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k         const                      iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
  ) = 
{
  tco_product_zero, tco_product<1>,  tco_product<2>,  tco_product<3>,  tco_product<4>,  tco_product<5>, tco_product<6>,
  tco_product<7>,   tco_product<8>,  tco_product<9>,  tco_product<10>, tco_product<11>, tco_product<12>,
  tco_product<13>,  tco_product<14>, tco_product<15>, tco_product<16>, tco_product<17>, tco_product<18>,
  tco_product<19>,  tco_product<20>, tco_product<21>, tco_product<22>, tco_product<23>, tco_product<24>, 
  tco_product<25>,  tco_product<26>, tco_product<27>, tco_product<28>, tco_product<29>, tco_product<30>,
  tco_product_final
};

template<const unsigned int N>
inline void
tco_product
(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k                                    iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
)
{
  const fld_k& xr = psi[iRow];
  fld_k tmp = 0.0;
  #pragma unroll (N)
  for (idx_k i=0; i<N; ++i)
  {
    tmp += upper[iFace]*psi[neigh[iFace]];
    Apsi[neigh[iFace]] += lower[iFace]*xr;
    ++iFace;
  }
  Apsi[iRow] += tmp;

  return tape[blkKind[++iRow]](blkKind, iFace, iRow, upper, lower, neigh, psi, Apsi);
};

inline void
tco_product_zero
(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k                                    iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
)
{
  return tape[blkKind[++iRow]](blkKind, iFace, iRow, upper, lower, neigh, psi, Apsi);
};

inline void
tco_product_final
(
  unsigned char const * const __restrict__ blkKind,
  idx_k&                                   iFace,
  idx_k                                    iRow,
  LDUoffDiag_k  const * const __restrict__ upper,
  LDUoffDiag_k  const * const __restrict__ lower,
  idx_k         const * const __restrict__ neigh,
  fld_k         const * const __restrict__ psi,
  fld_k               * const __restrict__ Apsi
)
{
  return;
};

void
matVec_TCO_CSC_CSR_CPU
(
  idx_k M, // = m_IO.matSize();
  unsigned char const * const  __restrict__ blkKind,
  idx_k         const * const  __restrict__ o, //=Permutation.pOwner();
  idx_k         const * const  __restrict__ n, //=Permutation.pNeigh();
  LDUoffDiag_k  const * const  __restrict__ u, //=Permutation.upper( randomMatrix.upper() );
  LDUoffDiag_k  const * const  __restrict__ l, //=Permutation.lower( randomMatrix.lower() );
  fld_k         const * const  __restrict__ x, //=Permutation.x(     randomMatrix.x()     );
  fld_k               * const  __restrict__ b  //=m_result;
)
{
  // Local variables
  idx_k iFace = 0;
  idx_k iRow = 0;

  tape[blkKind[0]](blkKind, iFace, iRow, u, l, n, x, b);
  return;
}
#endif

#ifdef __TEST_NATIVE_CUDA__
__global__
void
matVec_TCO_CSC_CSR_nativeCUDA
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#ifdef __TEST_OPEN_ACC__
void
matVec_TCO_CSC_CSR_openACC
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#ifdef __TEST_OPEN_MP__
void
matVec_TCO_CSC_CSR_openMP
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#endif
