#ifndef __MATVEC_padFastUnJ_CSC_CSR_KERNELS_H__
#define __MATVEC_padFastUnJ_CSC_CSR_KERNELS_H__

#define __MY_RESTRICT__ __restrict__

#ifdef __TEST_CPU__
inline void
fproductU
(
  idx_k        const                      nCols, 
  idx_k&                                  iFace,
  idx_k&                                  iRow,
  LDUoffDiag_k const * const __restrict__ upper,
  LDUoffDiag_k const * const __restrict__ lower,
  idx_k        const * const __restrict__ neigh,
  fld_k        const * const __restrict__ psi,
  fld_k              * const __restrict__ Apsi
)
{
  const fld_k& xr = psi[iRow];
  fld_k tmp = 0.0;
  for (idx_k i=0; i<nCols; ++i)
  {
    tmp += upper[iFace]*psi[neigh[iFace]];
    Apsi[neigh[iFace]] += lower[iFace]*xr;
    ++iFace;
  }
  Apsi[iRow++] += tmp;
  return;
};

template <const unsigned int M>
inline void
fproductU
(
  idx_k        const                      nCols, 
  idx_k&                                  iFace,
  idx_k&                                  iRow,
  LDUoffDiag_k const * const __restrict__ upper,
  LDUoffDiag_k const * const __restrict__ lower,
  idx_k        const * const __restrict__ neigh,
  fld_k        const * const __restrict__ psi,
  fld_k              * const __restrict__ Apsi
)
{
  double tmp[M];
  int facerows[M];

  double const* const __restrict__ psib = &psi[iRow];
  double* const __restrict__ Apsib = &Apsi[iRow];

  #ifdef __clang__
    /*code specific to clang compiler*/
  #elif __INTEL_COMPILER
    __assume_aligned(Apsib, 8*M);
  #elif __GNUC__
    __builtin_assume_aligned(Apsib, 8*M);
  #endif
  #pragma unroll_and_jam(M)
  for (int i=0; i<M; ++i)
  {
      tmp[i] = 0.;
      facerows[i] = iFace + i*nCols;
      for (int coli=0; coli<nCols; ++coli)
      {
          tmp[i] += upper[facerows[i]]*psi[neigh[facerows[i]]];
          Apsi[neigh[facerows[i]]] += lower[facerows[i]]*psib[i];
          facerows[i]++;
      }
      Apsib[i] += tmp[i];
  }
  iFace += M*nCols;
  iRow += M;
  return;
};

template <const unsigned int M, const unsigned int N>
inline void fproductF(idx_k& facei, idx_k& rowi,
                     LDUoffDiag_k const* const __restrict__ upper,
                     LDUoffDiag_k const* const __restrict__ lower,
                     idx_k const* const __restrict__ neigh,
                     fld_k const* const __restrict__ psi,
                     fld_k* const __restrict__ Apsi)
{
  double tmp[M];
  int facerows[M];

  double const* const __restrict__ psib = &psi[rowi];
  double* const __restrict__ Apsib = &Apsi[rowi];

  #ifdef __clang__
    /*code specific to clang compiler*/
  #elif __INTEL_COMPILER
    __assume_aligned(Apsib, 8*M);
  #elif __GNUC__
    __builtin_assume_aligned(Apsib, 8*M);
  #endif

  #pragma unroll_and_jam(M)
  for (int i=0; i<M; ++i)
  {
      tmp[i] = 0.;
      facerows[i] = facei + i*N;
      
      #pragma unroll(N)
      for (int coli=0; coli<N; ++coli)
      {
          tmp[i] += upper[facerows[i]]*psi[neigh[facerows[i]]];
          Apsi[neigh[facerows[i]]] += lower[facerows[i]]*psib[i];
          facerows[i]++;
      }

      Apsib[i] += tmp[i];
  }

  facei += M*N;
  rowi += M;
};


void
matVec_padFastUnJ_CSC_CSR_CPU
(
  idx_k nRows, // = m_IO.matSize();
  idx_k nBlks, 
  unsigned char const * const  __restrict__ nnz, // NOTE: this is the nnz for block
  idx_k         const * const  __restrict__ n, //=Permutation.pNeigh();
  LDUoffDiag_k  const * const  __restrict__ u, //=Permutation.upper( randomMatrix.upper() );
  LDUoffDiag_k  const * const  __restrict__ l, //=Permutation.lower( randomMatrix.lower() );
  fld_k         const * const  __restrict__ x, //=Permutation.x(     randomMatrix.x()     );
  fld_k               * const  __restrict__ b  //=m_result;
)
{
  // Local variables
  idx_k iFace = 0;
  idx_k iRow = 0;
  idx_k nnzLastBlk = 0;

  // Off diagonal part of the product
  for (int iBlk=0; iBlk<nBlks; ++iBlk)
  {
    switch ( nnz[iBlk] )
    {
      case (0):        iRow += 4; break;
      case (1):        fproductF<4, 1 >(iFace, iRow, u, l, n, x, b); break;
      case (2):        fproductF<4, 2 >(iFace, iRow, u, l, n, x, b); break;
      case (3):        fproductF<4, 3 >(iFace, iRow, u, l, n, x, b); break;
      case (4):        fproductF<4, 4 >(iFace, iRow, u, l, n, x, b); break;
      case (5):        fproductF<4, 5 >(iFace, iRow, u, l, n, x, b); break;
      case (6):        fproductF<4, 6 >(iFace, iRow, u, l, n, x, b); break;
      case (7):        fproductF<4, 7 >(iFace, iRow, u, l, n, x, b); break;
      case (8):        fproductF<4, 8 >(iFace, iRow, u, l, n, x, b); break;
      case (9):        fproductF<4, 9 >(iFace, iRow, u, l, n, x, b); break;
      case (10):       fproductF<4, 10>(iFace, iRow, u, l, n, x, b); break;
      case (11):       fproductF<4, 11>(iFace, iRow, u, l, n, x, b); break;
      case (12):       fproductF<4, 12>(iFace, iRow, u, l, n, x, b); break;
      case (13):       fproductF<4, 13>(iFace, iRow, u, l, n, x, b); break;
      case (14):       fproductF<4, 14>(iFace, iRow, u, l, n, x, b); break;
      case (15):       fproductF<4, 15>(iFace, iRow, u, l, n, x, b); break;
      case (16):       fproductF<4, 16>(iFace, iRow, u, l, n, x, b); break;
      case (17):       fproductF<4, 17>(iFace, iRow, u, l, n, x, b); break;
      case (18):       fproductF<4, 18>(iFace, iRow, u, l, n, x, b); break;
      case (19):       fproductF<4, 19>(iFace, iRow, u, l, n, x, b); break;
      case (20):       fproductF<4, 20>(iFace, iRow, u, l, n, x, b); break;
      case (21):       fproductF<4, 21>(iFace, iRow, u, l, n, x, b); break;
      default:         fproductU<4>(nnz[iBlk], iFace, iRow, u, l, n, x, b); break;
    }
  }

  nnzLastBlk = nnz[nBlks];
  while(iRow < nRows){
    fproductU(nnzLastBlk, iFace, iRow, u, l, n, x, b);
  }
  return;
};
#endif

#ifdef __TEST_NATIVE_CUDA__
__global__
void
matVec_padFastUnJ_CSC_CSR_nativeCUDA
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif


#ifdef __TEST_OPEN_ACC__
void
matVec_padFastUnJ_CSC_CSR_openACC
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#ifdef __TEST_OPEN_MP__
void
matVec_padFastUnJ_CSC_CSR_openMP
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#endif
