#ifndef __MATVEC_padFast_CSC_CSR_KERNELS_H__
#define __MATVEC_padFast_CSC_CSR_KERNELS_H__

#define __MY_RESTRICT__ __restrict__

#ifdef __TEST_CPU__
inline void
fproductF
(
  idx_k        const                      nCols, 
  idx_k&                                  iFace,
  idx_k&                                  iRow,
  LDUoffDiag_k const * const __restrict__ upper,
  LDUoffDiag_k const * const __restrict__ lower,
  idx_k        const * const __restrict__ neigh,
  fld_k        const * const __restrict__ psi,
  fld_k              * const __restrict__ Apsi
)
{
  const fld_k& xr = psi[iRow];
  fld_k tmp = 0.0;
  for (idx_k i=0; i<nCols; ++i)
  {
    tmp += upper[iFace]*psi[neigh[iFace]];
    Apsi[neigh[iFace]] += lower[iFace]*xr;
    ++iFace;
  }
  Apsi[iRow++] += tmp;
  return;
};


template<const unsigned int N>
inline void
fproductF
(
  idx_k&                                  iFace,
  idx_k&                                  iRow,
  LDUoffDiag_k const * const __restrict__ upper,
  LDUoffDiag_k const * const __restrict__ lower,
  idx_k        const * const __restrict__ neigh,
  fld_k        const * const __restrict__ psi,
  fld_k              * const __restrict__ Apsi
)
{
  const fld_k& xr = psi[iRow];
  fld_k tmp = 0.0;
  #pragma unroll (N)
  for (idx_k i=0; i<N; ++i)
  {
    tmp += upper[iFace]*psi[neigh[iFace]];
    Apsi[neigh[iFace]] += lower[iFace]*xr;
    ++iFace;
  }
  Apsi[iRow++] += tmp;
  
  return;
};


template<const unsigned int M>
void
matVec_padFast_CSC_CSR_CPU
(
  idx_k nRows, // = m_IO.matSize();
  idx_k nBlks, 
  unsigned char const * const  __restrict__ blkNNZ, // NOTE: this is the nnz for block
  idx_k         const * const  __restrict__ n, //=Permutation.pNeigh();
  LDUoffDiag_k  const * const  __restrict__ u, //=Permutation.upper( randomMatrix.upper() );
  LDUoffDiag_k  const * const  __restrict__ l, //=Permutation.lower( randomMatrix.lower() );
  fld_k         const * const  __restrict__ x, //=Permutation.x(     randomMatrix.x()     );
  fld_k               * const  __restrict__ b  //=m_result;
)
{
  // Local variables
  idx_k iFace = 0;
  idx_k iRow = 0;
  idx_k nnzLastBlk = 0;

  // Off diagonal part of the product
  for (int iBlk=0; iBlk<nBlks; ++iBlk)
  {
    switch ( blkNNZ[iBlk] )
    {
      case (0):        iRow += M; break;
      case (1):        for(idx_k j=0; j<M; j++){fproductF<1 >(iFace, iRow, u, l, n, x, b);} break;
      case (2):        for(idx_k j=0; j<M; j++){fproductF<2 >(iFace, iRow, u, l, n, x, b);} break;
      case (3):        for(idx_k j=0; j<M; j++){fproductF<3 >(iFace, iRow, u, l, n, x, b);} break;
      case (4):        for(idx_k j=0; j<M; j++){fproductF<4 >(iFace, iRow, u, l, n, x, b);} break;
      case (5):        for(idx_k j=0; j<M; j++){fproductF<5 >(iFace, iRow, u, l, n, x, b);} break;
      case (6):        for(idx_k j=0; j<M; j++){fproductF<6 >(iFace, iRow, u, l, n, x, b);} break;
      case (7):        for(idx_k j=0; j<M; j++){fproductF<7 >(iFace, iRow, u, l, n, x, b);} break;
      case (8):        for(idx_k j=0; j<M; j++){fproductF<8 >(iFace, iRow, u, l, n, x, b);} break;
      case (9):        for(idx_k j=0; j<M; j++){fproductF<9 >(iFace, iRow, u, l, n, x, b);} break;
      case (10):       for(idx_k j=0; j<M; j++){fproductF<10>(iFace, iRow, u, l, n, x, b);} break;
      case (11):       for(idx_k j=0; j<M; j++){fproductF<11>(iFace, iRow, u, l, n, x, b);} break;
      case (12):       for(idx_k j=0; j<M; j++){fproductF<12>(iFace, iRow, u, l, n, x, b);} break;
      case (13):       for(idx_k j=0; j<M; j++){fproductF<13>(iFace, iRow, u, l, n, x, b);} break;
      case (14):       for(idx_k j=0; j<M; j++){fproductF<14>(iFace, iRow, u, l, n, x, b);} break;
      case (15):       for(idx_k j=0; j<M; j++){fproductF<15>(iFace, iRow, u, l, n, x, b);} break;
      case (16):       for(idx_k j=0; j<M; j++){fproductF<16>(iFace, iRow, u, l, n, x, b);} break;
      case (17):       for(idx_k j=0; j<M; j++){fproductF<17>(iFace, iRow, u, l, n, x, b);} break;
      case (18):       for(idx_k j=0; j<M; j++){fproductF<18>(iFace, iRow, u, l, n, x, b);} break;
      case (19):       for(idx_k j=0; j<M; j++){fproductF<19>(iFace, iRow, u, l, n, x, b);} break;
      case (20):       for(idx_k j=0; j<M; j++){fproductF<20>(iFace, iRow, u, l, n, x, b);} break;
      case (21):       for(idx_k j=0; j<M; j++){fproductF<21>(iFace, iRow, u, l, n, x, b);} break;
      default:         for(idx_k j=0; j<M; j++){fproductF(blkNNZ[iBlk], iFace, iRow, u, l, n, x, b);} break;
    }
  }

  nnzLastBlk = blkNNZ[nBlks];
  while(iRow < nRows){
    fproductF(nnzLastBlk, iFace, iRow, u, l, n, x, b);
  }
  
  return;
};
#endif

#ifdef __TEST_NATIVE_CUDA__
__global__
void
matVec_padFast_CSC_CSR_nativeCUDA
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif


#ifdef __TEST_OPEN_ACC__
void
matVec_padFast_CSC_CSR_openACC
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#ifdef __TEST_OPEN_MP__
void
matVec_padFast_CSC_CSR_openMP
(
    const idx_k                               nCells,
    const idx_k                               nFaces,
    const idx_k                               nInternalFaces,
    const idx_k                               cellsStride,
    const idx_k                               facesStride,
    const idx_k                               iComp,
    const idx_k*        const __MY_RESTRICT__ owner,
    const idx_k*        const __MY_RESTRICT__ neigh,
    const LDUoffDiag_k* const __MY_RESTRICT__ upper,
    const LDUoffDiag_k* const __MY_RESTRICT__ lower,
    const fld_k*        const __MY_RESTRICT__ psi,
          fld_k*        const __MY_RESTRICT__ Apsi
)
{
  return;
};
#endif

#endif
