# Utility to join databases generated from matVecKernel test.
# default: join all the 

import os
import argparse
import sqlite3

def joinDb(name, listDb, tableData, nValues):
    valueStr = ""
    for i in range(0, nValues-1):
        valueStr += "?, "
    valueStr += "?"

    dbBase = sqlite3.connect(name)
    dbBase_cursor = dbBase.cursor()

    dbBase_cursor.execute(f"CREATE TABLE IF NOT EXISTS profiling ({tableData})")

    for db in listDb:
        print(db)
        dbToAppend = sqlite3.connect(db)
        dbToAppend_cursor = dbToAppend.cursor()
        dbToAppend_cursor.execute(f'SELECT {tableData} FROM profiling')
        output = dbToAppend_cursor.fetchall()

        for row in output:
            dbBase_cursor.execute(f"INSERT INTO profiling VALUES ({valueStr})", row)
        dbBase.commit()
        dbToAppend_cursor.close()
    
    dbBase_cursor.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='unionDb', allow_abbrev=False)
    parser.add_argument("path", nargs=1, type=str, help="base path for the research")
    parser.add_argument("--databases", nargs="*", type=str, required=False, help="Databases to be merged (optional)")
    parser.add_argument("--cluster", nargs="*", type=str, required=False, help="Cluster on which the run where made")
    parser.add_argument("--instrumentation", nargs="?", type=str, required=False, 
                        help='''Specific instrumentation to join. If 'baseMetrics',
                          all datatBases will be joined consodering only the base metrices.''')
    args = parser.parse_args()

    basePath = vars(args)['path'][0]
    cluster = vars(args)['cluster']
    if cluster:
        cluster = cluster[0]
    databases = vars(args)['databases']
    instrumentation = vars(args)['instrumentation']

    # Verify id the user has defined a database list
    # othrwise take all the .db in the basepath
    if not databases:
        print("No databases specified...")
        databases = []
        mats = os.listdir(basePath)
        for mat in mats:
            # traverse whole directory
            for root, dirs, files in os.walk(os.path.join(basePath, mat)):
                # select file name
                for file in files:
                    # check the extension of files
                    if file.endswith('.sqlite.db') and "BKP" not in file and "BKP" not in root:
                        databases.append(os.path.join(root, file))
    
    print(f"Total number of databases to be managed: {len(databases)}")

    # filter databases according to instrumentation:
    dbStlOps = list(filter(lambda x: "instrStlOps" in x, databases))
    dbCache = list(filter(lambda x: "instrCache" in x, databases))
    dbIns = list(filter(lambda x: "instrIns" in x, databases))
    dbOpsIns = list(filter(lambda x: "instrOpsIns" in x, databases)) # DGX
    dpCacheOps = list(filter(lambda x: "instrCacheOps" in x, databases)) #E4
    dbOps = list(filter(lambda x: "instrOps" in x, databases)) #Leonardo
    dbLST = list(filter(lambda x: "instrLST" in x, databases)) #Leonardo
    dbFP = list(filter(lambda x: "instrFP" in x, databases)) #Leonardo
    dbNoInst = list(filter(lambda x: "instr" not in x, databases))
    dbBaseMetrics = list(filter(lambda x: "BaseMetrix" not in x, databases))
    if len(dbBaseMetrics) == 0:
        dbBaseMetrics = databases

    if dbStlOps and (instrumentation == "stlOps" or not instrumentation):
        print(f"n databases for stlOps: {len(dbStlOps)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
            stallMemWrite double, stallRes double, totalCycles double, dpOperation double, dpVectorOperations double"
        nValues = 28
        joinDb(os.path.join(basePath, "dbStlOps.sqlite.db"), dbStlOps, dataNames, nValues)

    if dbCache and (instrumentation == "cache" or not instrumentation):
        print(f"n databases for cache: {len(dbCache)}")   
        if cluster == "dgx":
            dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
                nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
                nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
                warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
                L1_ICM double, L2_DCM double, L2_ICM double"
            nValues = 26
        else:
            dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
                nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
                nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
                warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
                L1_DCM double, L1_ICM double, L2_DCM double, L2_ICM double, L3_TCM double"
            nValues = 28
        joinDb(os.path.join(basePath, "dbCache.sqlite.db"), dbCache, dataNames, nValues)

    if dbIns and (instrumentation == "ins" or not instrumentation):
        print(f"n databases for ins: {len(dbIns)}")
        if cluster == "e4":
            dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
                nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
                nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
                warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
                L2_ICM double, branchInstraction double, branchCondIns double, branchMissPred double, \
                totalInstructions double"
            nValues = 28
        else:        
            dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
                nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
                nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
                warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
                loadInstructions double, storeInstruction double, branchInstraction double, branchCondIns double, \
                branchMissPred double, totalInstructions double"
            nValues = 29
        joinDb(os.path.join(basePath, "dbIns.sqlite.db"), dbIns, dataNames, nValues)

    if dbOpsIns and (instrumentation == "opsIns" or not instrumentation):
        print(f"n databases for opsIns: {len(dbOpsIns)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
            totalCycles double, floatingPointOps double, branchInstraction double, branchMissPred double, \
            totalInstructions double"
        nValues = 28
        joinDb(os.path.join(basePath, "dbOpsIns.sqlite.db"), dbOpsIns, dataNames, nValues)

    if dbOps and (instrumentation == "ops" or not instrumentation):
        print(f"n databases for opsIns: {len(dbOps)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, totalCycles double, \
            dpOperation double, dpVectorOperations double"
        nValues = 26
        joinDb(os.path.join(basePath, "dbOps.sqlite.db"), dbOps, dataNames, nValues)

    if dbLST and (instrumentation == "LST" or not instrumentation):
        print(f"n databases for LST: {len(dbLST)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, loadStoredInstructions double"
        nValues = 24
        joinDb(os.path.join(basePath, "dbLST.sqlite.db"), dbLST, dataNames, nValues)

    if dbFP and (instrumentation == "FP" or not instrumentation):
        print(f"n databases for FP: {len(dbFP)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, fpInstructions double"
        nValues = 24
        joinDb(os.path.join(basePath, "dbFP.sqlite.db"), dbFP, dataNames, nValues)

    if dpCacheOps and instrumentation == "cacheOps" or not instrumentation:
        print(f"n databases for cacheOps: {len(dpCacheOps)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double, \
            L1_DCM double, L2_DCM double, fpOperation double, vectorIns double, totalCycles double"
        nValues = 28
        joinDb(os.path.join(basePath, "dpCacheOps.sqlite.db"), dpCacheOps, dataNames, nValues)

    if dbNoInst and instrumentation == "noInstr" or not instrumentation:
        print(f"n databases for noInstr: {len(dbNoInst)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double"
        nValues = 23
        joinDb(os.path.join(basePath, "dbNoInst.sqlite.db"), dbNoInst, dataNames, nValues)

    if instrumentation == "baseMetrics":
        print(f"n databases for baseMetrics: {len(dbBaseMetrics)}")
        dataNames = "date text, machine text, compiler text, flags text, hostname text, prefix text, isRCM int, \
            nRows int, procID int, kernel text, kernelFlavour text, caseIndex int, mpiIndex int, mpiSize int, \
            nRepetitions int, bandWidth double, jumpParam double, warmupElapsedTime double, warmupClockTime double, \
            warmupCudaTime double, raceElapsedTime double, raceClockTime double, raceCudaTime double"
        nValues = 23
        joinDb(os.path.join(basePath, "dbBaseMetrics.sqlite.db"), dbBaseMetrics, dataNames, nValues)


