import sqlite3
import os
import numpy as np
import matplotlib.pyplot as plt

# DataBase TABLE CONTENT
# NAME: profiling
# 1.  date text, 
# 2.  machine text, 
# 3.  compiler text, 
# 4.  flags text, 
# 5.  hostname text, 
# 6.  prefix text, 
# 7.  isRCM int, 
# 8.  nRows int,
# 9.  procID int,
# 10. kernel text,
# 11. kernelFlavour text,
# 12. caseIndex int,
# 13. mpiIndex int,
# 14. mpiSize int,
# 16. nRepetitions int,
# 17. bandWidth double,
# 18. jumpParam double,
# 19. warmupElapsedTime double,
# 20. warmupClockTime double,
# 21. warmupCudaTime double,
# 22. raceElapsedTime double,   (mean value over nRepetition runs)
# 23. raceClockTime double,     (mean value over nRepetition runs)
# 24. raceCudaTime double,      (mean value over nRepetition runs)
# 25. cacheMisses double,       (mean value over nRepetition runs)
# 26. branchMisses double       (mean value over nRepetition runs)



def sortingFunc(e):
    value = 0
    if "BlockRowPerm" in e:
        value = int(e.split("_")[2].split(".")[0])
    return value

def plotGraph(quantity, quantityType, path, std = False):
    fig, ax = plt.subplots()
    if std:
        ax.bar(list(quantity.keys()), list(quantity.values()), width=0.5, yerr=np.std(list(quantity.values())))
    else:
        ax.bar(list(quantity.keys()), list(quantity.values()), width=0.5)
    plt.xticks(range(len(quantity)), list(quantity.keys()), rotation='vertical')
    ax.set_ylabel(f"{quantityType} [s]")
    ax.set_title(f"{mat} - {nRow}: {quantityType} comparison ")
    fig.savefig(os.path.join(path, f"{mat}_{nRow}-{quantityType}_comparison.png"), bbox_inches='tight')

if __name__ == '__main__':
    dataPath = "/m100_work/ExaF_prod23/fbottau0/matveckernels/data"
    matsType = os.listdir(dataPath)
    #matsType = ["PLMmats-RCM", "PLMmats_CFmesh-RCM"]

    raceTimeMaxMean = dict()    # Dictionary of the mean of between the maximum race time used by the MPIprocess
    raceTimeMaxMax = dict()     # Dictionary of the maximum of between the maximum race time used by the MPIprocess
    warmUpTimeMaxMean = dict()  # Dictionary of the mean of between the maximum warm up time used by the MPIprocess
    warmUpTimeMaxMax = dict()   # Dictionary of the maximum of between the maximum warm up time used by the MPIprocess
    cacheMisses = dict()
    branchMisses = dict()
    
    for mat in matsType:
        print(f"processing matrix type: {mat}")
        matPath = os.path.join(dataPath, mat)
        #nRows = list(filter(lambda dir: "run" not in dir, os.listdir(matPath)))
        nRows =["10krows", "50krows", "90krows"]

        for nRow in nRows:
            print(f"processing number of rows: {nRow}")
            nRowPath = os.path.join(matPath, nRow)
            dataBases = list(filter(lambda file: ".db" in file, os.listdir(nRowPath)))
            dataBases.sort(key=sortingFunc)
            for db in dataBases:
                alg = db.replace("timings_", "").strip(".sqlite.db")
                print(f"Processing DataBase: {db}, Algorithm = {alg}")
                table = sqlite3.connect(os.path.join(nRowPath, db))
                cur = table.cursor()
                dataSet = list(cur.execute("SELECT procID, raceClockTime, warmupClockTime, cacheMisses, branchMisses FROM profiling"))
                rTimeTmp = []
                wuTimeTmp = []
                raceTime = []
                wuTime = []
                cacheMiss = []
                branchMiss = []
                procID = dataSet[0][0]
                print(f"processing processor{procID}")
                for i in range(0, len(dataSet)):
                    cacheMiss.append(dataSet[i][3])
                    branchMiss.append(dataSet[i][4])
                    if dataSet[i][0] == procID:
                        rTimeTmp.append(dataSet[i][1])
                        wuTimeTmp.append(dataSet[i][2])
                    else:
                        raceTime.append(max(rTimeTmp))
                        wuTime.append(max(wuTimeTmp))
                        timeTmp = []
                        wuTimeTmp = []
                        procID = dataSet[i+1][0]
                        print(f"processing processor{procID}")
                raceTimeMaxMean[alg] = np.mean(raceTime)
                raceTimeMaxMax[alg] =  max(raceTime)
                warmUpTimeMaxMean[alg] = np.mean(wuTime)
                warmUpTimeMaxMax[alg] = max(wuTime)
                cacheMisses[alg] = np.mean(cacheMiss)
                branchMisses[alg] = np.mean(branchMiss)
                table.close()
            # plotting image
            print("Plotting images ...")
            plotGraph(raceTimeMaxMean, "Race Time Max Mean", nRowPath, std=True)
            plotGraph(raceTimeMaxMax, "Race Time Max Max", nRowPath)
            plotGraph(warmUpTimeMaxMean, "Warm Up Time Max Mean", nRowPath, std=True)
            plotGraph(warmUpTimeMaxMax, "Warm Up Time Max Max", nRowPath)
            plotGraph(cacheMisses, "Cache Misses", nRowPath)
            plotGraph(branchMisses, "Branch Misses", nRowPath)




                
                


