import sqlite3
import numpy as np
import argparse
import matplotlib.pyplot as plt
import os

def sortingFunc(e):
    value = 0
    if "BlockRowPerm" in e:
        value = int(e.split("_")[1]) #.split(".")[0])
    return value

def createDataDict(data):
    dataDict = dict()
    if timeAnalysis == "MeanMax":
        for datum in data:
            if not datum[0] in dataDict.keys():
                dataDict[datum[0]] = dict() # dictionary for each group
            if not datum[1] in dataDict[datum[0]].keys():
                dataDict[datum[0]][datum[1]] = dict() # dictionary for each type of comparison
            if not datum[2] in dataDict[datum[0]][datum[1]].keys():
                dataDict[datum[0]][datum[1]][datum[2]] = [] # vector of measurments relative to matrix decomposition
            dataDict[datum[0]][datum[1]][datum[2]].append(datum[3])
    else: 
        for datum in data:
            if not datum[0] in dataDict.keys():
                dataDict[datum[0]] = dict() # dictionary for each matrix format
            if not datum[1] in dataDict[datum[0]].keys():
                dataDict[datum[0]][datum[1]] = [] # vector with relative metrix to the object type
            dataDict[datum[0]][datum[1]].append(datum[2])

    return dataDict

def plotGraph(dataDict, compObj, types, figTlt, figNm):
    # -Plotting data
    x = []

    groups = list(dataDict.keys())
    groups.sort()
    multiplier = 0
    barWidth = 1/(len(groups)+1)
    
    fig1, ax1 = plt.subplots(figsize=[12, 4])
    ax1.set_ylabel(compObj)
    ax1.set_title(figTlt)
        
    for group in groups:
        x = np.arange(len(types)) #len(dataDict[group]))
        if timeAnalysis == "MaxMax":
            values = [max(dataDict[group][type]) for type in types]
            error = [np.zeros(len(types)), np.zeros(len(types))]

        elif timeAnalysis == "MeanMax":
            errorMin = []
            errorMax = []
            values = []
            for type in types:
                tmp = []
                for procID in dataDict[group][type]:
                    tmp.append(max(dataDict[group][type][procID]))
                meanTmp = np.mean(tmp)
                tmp = list(filter(lambda x: abs(x-meanTmp)<0.75*meanTmp, tmp))
                values.append(np.mean(tmp))
                errorMin.append(np.mean(tmp)-min(tmp, default=0))
                errorMax.append(max(tmp, default=0)-np.mean(tmp))
            error = [errorMin, errorMax]

        else:
            values = [np.mean(dataDict[group][flavor]) for flavor in types]
            error = [np.zeros(len(types)), np.zeros(len(types))]

        ax1.bar(x + multiplier*barWidth, values, width=barWidth, yerr=error, label=group)
        multiplier += 1
    
    ax1.set_xticks(x + barWidth*int(len(groups)/2))
    ax1.set_xticklabels(types, rotation=90)
    ax1.grid(axis='y')
    ax1.legend()
    fig1.savefig(os.path.join(os.path.dirname(db), f"{figNm}.png"), bbox_inches='tight') 
    plt.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='sqliteAnalysis', allow_abbrev=False)
    parser.add_argument('dataBase', nargs=1, type=str, help="database to be analysed")
    args = parser.parse_args()

    db = vars(args)['dataBase'][0]
    
    dbConnect = sqlite3.connect(db)
    dbCursor = dbConnect.cursor()
    data = dbCursor.execute("SELECT * FROM profiling")

    #- Choose whether a graph or a numeric table is needed
    outputFormat = input(f"Choose output format (graph, table): \n")
    print(f"You choose: {outputFormat}")

    #- Display availavble data:
    print(f"Data contained in the dataBase:")
    for column in data.description:
        print(column[0])
    print()
    
    if outputFormat == "graph":
        #- Choose what to compare
        comparisonObject = input("Type what to be compared (Y axis of the graph) form the data in the table: \n")
        print(f"You chose: {comparisonObject}\n")

        comparisonType = input("Type against what make the comparison (X axis of the graph): kernelFlavour, compiler, nRows ... : \n")
        print(f"You chose: {comparisonType}\n")
        if comparisonType == "meshType":
            comparisonType = "prefix"
        
        comparisonGroup = input("Type groups of data to consider (e.g. meshType, compiler, nRows ...)\n")
        print(f"You chose: {comparisonGroup}\n")
        if comparisonGroup == "meshType":
            comparisonGroup = "prefix"

        compiler = ""
        if comparisonType != "compiler" and comparisonGroup !="compiler":
            compilerBool = input("Do you whant to specify a compiler (yes or no)? ")
            if compilerBool == "yes":
                compilerAv = dbCursor.execute("SELECT DISTINCT compiler FROM profiling")
                print(f"available compilers:")
                for comp in compilerAv:
                    print(comp[0])
                compiler = input("Choose which compiler to consider: ")
                print(f"You chose: {compiler}\n")

        date = ""
        dateBool = input("Do you whant to specify a test date (yes or no)? ")
        if dateBool == "yes":
            dateAv = dbCursor.execute("SELECT DISTINCT date FROM profiling")
            print(f"available test dates:")
            for date in dateAv:
                print(date[0])
            date = input('Choose which date to consider (between " "): ')
            print(f"You chose: {date}\n")

        mpiSize = ""
        mpiSizeBool = input("Do you whant to specify an MPI size (yes or no)? ")
        if mpiSizeBool == "yes":
            mpiSizeAv = dbCursor.execute("SELECT DISTINCT mpiSize FROM profiling")
            print(f"available MPI sizes :")
            for mpisize in mpiSizeAv:
                print(mpisize[0])
            mpiSize = input('Choose which MPI size to consider: ')
            print(f"You chose: {mpiSize}\n")

        kernels= ""
        if comparisonType == "kernelFlavour" or comparisonGroup =="kernelFlavour":
            kernelsBool = input("Do you whant to specify a kernelFlavour (yes or no)? ")
            if kernelsBool == "yes":
                kernelsAv = dbCursor.execute("SELECT DISTINCT kernelFlavour FROM profiling")
                print(f"available kernelFlavours:")
                for kernel in kernelsAv:
                    print(kernel[0])
                kernels = input("Type kernels to be analysed (comma separated if more than 1): \n")
                kernels = kernels.split(',')
                nKernels = len(kernels)
                print(f"You chose {nKernels} kernels: {kernels}\n")


        timeAnalysis = ""
        if "Time" in comparisonObject:
            timeAnalysis = input("Choose how to analyse time: MeanMax or MaxMax \n")
            print(f"Time analysis type: {timeAnalysis}\n")

        nRows = []
        if comparisonType != "nRows" and comparisonGroup !="nRows":
            nRowsAv = dbCursor.execute("SELECT DISTINCT nRows FROM profiling")
            nRowsAv = list(nRowsAv)
            nRowsAv.sort()
            nRowsAv = [x[0] for x in nRowsAv]
            specificNrows = input("Do you want to select a specific matrix dimention (yes or no)? ")
            if specificNrows == "yes":
                print(f"available matrix dimensions:")
                for nrows in nRowsAv:
                    print(nrows)
                nRows = [int(input("Insert the desired matrix dimension: "))]
            else:
                combinatedNRows = input("Do you want to make separated plot for each matrix dimentions (yes or no)? ")
                if combinatedNRows == "yes":
                    nRows = nRowsAv
                
        
        #- Data extraction from database and plotting
        
        if timeAnalysis == "MeanMax":
            columns = f"{comparisonGroup}, {comparisonType}, procID, {comparisonObject}"
        else:
            columns = f"{comparisonGroup}, {comparisonType}, {comparisonObject}"

        if comparisonType == "prefix":  comparisonType = "meshType"
        if comparisonObject == "prefix":  comparisonObject = "meshType"
        
        if nRows != []:
            for nrows in nRows:
                print(f"Analysing data for {nrows}")

                dbQuery = F"SELECT {columns} FROM profiling WHERE nRows={nrows}"

                if compiler: dbQuery = dbQuery + f" AND compiler='{compiler}'"
                if date: dbQuery = dbQuery + f" AND date='{date}'"
                if kernels:
                    dbQuery = dbQuery + " AND ("
                    for kernel in kernels:
                        dbQuery = dbQuery + f"kernelFlavour='{kernel}' OR "
                    dbQuery = dbQuery.strip(' OR ') + ")"
                if mpiSize: dbQuery = dbQuery + f" AND mpiSize = {mpiSize}"

                data = list(dbCursor.execute(dbQuery))

                types = list(set([x[1] for x in data]))
                types.sort()
                types.sort(key = sortingFunc)

                #- Extracting and grouping data
                dataDict = createDataDict(data)

                # -Plotting data
                figTitle = f"{nrows} {compiler}"

                figName = f"{nrows}-{comparisonObject}{timeAnalysis}-VS-{comparisonType}_{comparisonGroup}"
                if compiler != "":
                    compiler.strip("\'").strip("\'")
                    figName += f"-{compiler}"

                plotGraph(dataDict, comparisonObject, types, figTitle, figName)
        
        else:
            print("Analysing data...")

            nRowsAv = dbCursor.execute("SELECT DISTINCT nRows FROM profiling")
            nRows = list(nRowsAv)
            nRows.sort()
            nRows = [x[0] for x in nRows]

            #- select only available kernelFlavour
            typesAv = dbCursor.execute(f"SELECT DISTINCT kernelFlavour FROM profiling WHERE nRows={nRows[0]}")
            types = [x[0] for x in data]
            types.sort()
            types.sort(key = sortingFunc)

            dbQuery = F"SELECT {columns} FROM profiling"

            if compiler: dbQuery = dbQuery + f" AND compiler='{compiler}'"
            if date: dbQuery = dbQuery + f" AND date='{date}'"
            if kernels:
                dbQuery = dbQuery + " AND ("
                for kernel in kernels:
                    dbQuery = dbQuery + f"kernelFlavour='{kernel}' OR "
                dbQuery = dbQuery.strip(' OR ') + ")"

            data = list(dbCursor.execute(dbQuery))
            
            #- Extracting and grouping data
            dataDict = createDataDict(data)

            # -Plotting data
            figTitle = f"{compiler}"

            figName = f"{comparisonObject}{timeAnalysis}-VS-{comparisonType}_{comparisonGroup}"
            if compiler != "":
                figName += f"-{compiler}"

            plotGraph(dataDict, comparisonObject, types, figTitle, figName)

    else: #- Output format Table       
        #- Choose what to compare
        metrics = input("Type metrics to be inserted in the table (comma separated if more than 1): \n")
        metricsLst = metrics.split(',')
        nMetrics = len(metricsLst)
        print(f"You chose {nMetrics}: {metrics}\n")

        #- Coose file name
        fileName = input("Type file name: ")
        fileName = fileName + ".csv"
        print(f"File name: {fileName}")

        nRowsAv = dbCursor.execute("SELECT DISTINCT nRows FROM profiling")
        nRows = list(nRowsAv)
        nRows.sort()
        nRows = [x[0] for x in nRows]
        print(f"Number of rows present in the dataBase: {nRows}")

        compilerAv = dbCursor.execute("SELECT DISTINCT compiler FROM profiling")
        compilers = list(compilerAv)
        compilers = [x[0] for x in compilers]
        print(f"Compilers present in the dataBase: {compilers}")

        meshTypesAv = dbCursor.execute("SELECT DISTINCT prefix FROM profiling")
        meshTypes = list(meshTypesAv)
        meshTypes = [x[0] for x in meshTypes]
        print(f"MeshType present in the dataBase: {meshTypes}")

        kernelsAv = dbCursor.execute("SELECT DISTINCT kernelFlavour FROM profiling")
        kernels = list(kernelsAv)
        kernels = [x[0] for x in kernels]
        kernels.sort(key = sortingFunc)
        print(f" spMVM kernels present in the dataBase: {kernels}")

        headerLine = f"nRows,meshType,kernel,compiler,{metrics} \n"

        f = open(fileName, "w")
        f.writelines(headerLine)

        for nRow in nRows:
            print(f"Analizing {nRow}")
            for meshType in meshTypes:
                print(f"\tAnalizing {meshType}")
                for kernel in kernels:
                    print(f"\t\tAnalizing {kernel}")
                    for compiler in compilers:
                        print(f"\t\t\tAnalizing {compiler}")
                        data = list(dbCursor.execute(f'''SELECT procID, {metrics} FROM profiling WHERE 
                                                     nRows={nRow} AND prefix='{meshType}' AND 
                                                     kernelFlavour='{kernel}' AND compiler='{compiler}' '''))
                        dataStr = f"{nRow},{meshType},{kernel},{compiler},"

                        for i in range(0, nMetrics):
                            value = 0
                            dataDict = dict()
                            if "Time" in metricsLst[i]:
                                for datum in data:
                                    if datum[0] not in dataDict.keys():
                                        dataDict[datum[0]] = []
                                    dataDict[datum[0]].append(datum[i+1])
                                for key in dataDict.keys():
                                    meanValue = np.mean(dataDict[key]) 
                                    if meanValue > value:
                                        value = meanValue
                            
                            else:
                                for datum in data:
                                    if datum[i+1] > value: value = datum[i+1]
                            
                            dataStr = dataStr + f"{value},"
                        
                        f.writelines(dataStr.rstrip(",")+"\n")
                        




        
