#!/usr/bin/python3

import os
import argparse

parser = argparse.ArgumentParser(prog="submitAllMats", allow_abbrev=False)
parser.add_argument("--specificMat", type=str, required=False, help="Specic matrix type to be tested")
parser.add_argument("--specificSize", type=str, required=False, help ="Specific nRows to be tested (indicated as XXkrows)")
parser.add_argument("--instrumented", type=str, default=None, help="stlOps: stalls and dp operation measurements, cache: cache instrumentation, ins: instruction measurements")
parser.add_argument("--loopType", type=int, default=0, help="0: fixedMatType, 1:fixedNrows")
args = parser.parse_args()

specMat = vars(args)['specificMat']
specSize = vars(args)['specificSize']
instr = vars(args)['instrumented']
loopType = vars(args)['loopType']

basePath = os.path.abspath(os.path.join(os.getcwd(), os.pardir))

testMatPath = os.path.join(basePath, "data")
jobScriptPath = os.getcwd()
binPath = os.path.join(basePath, "matVecKernels/bin")

if instr == "stlOps":
    binPath = os.path.join(binPath, "matVecKernels_instrStlOps.x")
    instrFlag = "instrStlOps"
elif instr == "cache":
    binPath = os.path.join(binPath, "matVecKernels_instrCache.x")
    instrFlag = "instrCache"
elif instr == "ins":
    binPath = os.path.join(binPath, "matVecKernels_instrIns.x")
    instrFlag = "instrIns"
else:
    print("Run not instrumented")
    binPath = os.path.join(binPath, "matVecKernels.x")
    instrFlag = "notInstr"

if specMat:
    matTypes = [specMat]
else:
    matTypes = os.listdir(testMatPath)

if specSize:
    nRows = [specSize]
else:
    nRows = ["10krows", "50krows", "90krows"]
    # nRows = os.listdir(os.path.join(testMatPath, matTypes[0]))

os.chdir(jobScriptPath)

with open("../matVecKernels/compileEnv_g100", "r") as f:
    for line in f.readlines():
        if "NCORE_PER_NODE" in line:
            ncore_per_node = int(line.split("=")[1])

if loopType == 0:
    for mat in matTypes:
        matPath = os.path.join(testMatPath, mat)
        command = f"sbatch --job-name {mat}-{instrFlag} --ntasks-per-node {ncore_per_node} {jobScriptPath}/submitMatVec_fixedMatType.sh {mat} {matPath} {binPath} {instrFlag}"
        print(command)
        os.system(command)

elif loopType == 1:
    for size in nRows:
        command = f"sbatch --job-name {size}-{instrFlag} --ntasks-per-node {ncore_per_node} {jobScriptPath}/submitMatVec_fixedNrows.sh {size} {testMatPath} {binPath} {instrFlag}"
        print(command)
        os.system(command)

else:
    print("Indicated loop type non correct: allowed 0 or 1")
