#!/bin/bash
#SBATCH --nodes=1
#SBATCH --ntasks-per-node=1
#SBATCH --gpus-per-node=0
#SBATCH --cpus-per-task=1
#SBATCH --output=R-%x.%j.out
#SBATCH --error=R-%x.%j.err
#SBATCH --account=ExaF_prod23_2
#SBATCH --partition=g100_usr_prod
#SBATCH --time=01:00:00
#SBATCH --qos=g100_qos_dbg

set -e

# User defined variables
NUM_MAT=1
NUM_VEC=1
CASE_TO_COMPUTE=("SHMmats-RCM-BRP") #$(ls /g100_work/ExaF_prod23_2/matVec_tests/TestMatrices | grep "BRP")

# Internal variables 
if [ -n $SLURM_JOB_ID ] ; then
  SCRIPT_PATH=$(scontrol show job $SLURM_JOBID | awk -F= '/Command=/{print $2}')
else
  SCRIPT_PATH="$( cd -- "$( dirname -- "${BASH_SOURCE[0]:-$0}"; )" &> /dev/null && pwd 2> /dev/null; )";
fi
ROOT_PATH=$(dirname $(dirname ${SCRIPT_PATH}))
DATA_DIR=${ROOT_PATH}"/data"
BIN_DIR=${ROOT_PATH}"/preProcessor/bin"
TEST_MATRICES_DIR=$(dirname ${ROOT_PATH})"/TestMatrices/"

#
# Clean environment
module purge

#
# Load needed environment
module load profile/global
module load git
module load cmake

# loading gcc modules 
module load gcc/10.2.0
module load openmpi/4.1.1--gcc--10.2.0-cuda-11.5.0 
module load zlib/1.2.11--gcc--10.2.0 
module load hdf5/1.10.7--openmpi--4.1.1--gcc--10.2.0


for mat in ${CASE_TO_COMPUTE[@]} ; do
  for filename in `find $TEST_MATRICES_DIR -type f -iname 'owner' | grep $mat` ; do
    relName=$(realpath --relative-to=${TEST_MATRICES_DIR} ${filename}) 
    relPath=$(dirname $(dirname ${relName}))
    prefix=`echo ${relName} | sed -e '{s/^\///};{s/\/owner//};{s/\//_/g}'`
    ownerName=${filename}
    neighName=`echo ${filename} | sed -e 's/owner/neighbour/'`;
    outDir=${DATA_DIR}"/"${relPath}
    mkdir -p ${outDir}
    ${BIN_DIR}"/FOAM2H5MatConverter" -o ${ownerName} -n ${neighName} -p ${prefix} -d ${outDir} -r ${NUM_MAT} -t ${NUM_VEC}
  done
done

