#!/bin/bash
#SBATCH --nodes=1
##SBATCH --ntasks-per-node=48
#SBATCH --gpus-per-node=0
#SBATCH --exclusive
#SBATCH --account=cin_staff
#SBATCH --partition=g100_usr_prod
#SBATCH --time 10:00:00
##SBATCH	--qos=m100_qos_dbg

# CL aurguments:
# {1}: matrix type
# {2}: matrices path
# {3}: executables path
# {4}: instrumentation flag
#
# Load the environment needed to run the job
exe=${3}
mkdir ${2}
cd ${2}

# User defined variables
#algorithms=("COO_COO"  "CSC_CSR"  "CSR_CSR"  "CSR"  "BlockRowPerm_5" "BlockRowPerm_10" "BlockRowPerm_15" "BlockRowPerm_20" "BlockRowPerm_25" "BlockRowPerm_30" "BlockRowPerm_35" "BlockRowPerm_40" "BlockRowPerm_45"  "BlockRowPerm_50" "BlockRowPerm_55" "BlockRowPerm_60" "BlockRowPerm_65" "BlockRowPerm_70" "BlockRowPerm_75" "BlockRowPerm_80" "BlockRowPerm_85" "BlockRowPerm_90" "BlockRowPerm_95" "BlockRowPerm_100"  )
algorithms=("COO_COO"  "CSC_CSR"  "CSR_CSR"  "CSR" "FAST_CSC_CSR" "TCO_CSC_CSR" "GTO_CSC_CSR" "padFast4_CSC_CSR" "padFast8_CSC_CSR" "padFastUnJ_CSC_CSR")
BLKblocks=("10" "50" "100" "250" "500" "1000" "2000" "3000" "5000" "7500" "10000" "12500")
if ! [[ ${1} == *"RCM"* ]]; then
	#algorithms=("COO_COO"  "CSC_CSR"  "CSR_CSR"  "CSR")
	BLKblocks=()
fi
minNrows=8
suffix="krows"

#
# Init log file 
date > ${2}/run_${1}-${4}.log

#
# Load PAPI environment
export PAPI_DIR="/m100/home/userinternal/fbottau0/papi/src/install/"
export PATH="$PAPI_DIR/bin:$PATH"
export LD_LIBRARY_PATH="$PAPI_DIR/lib:$LD_LYBRARY_PATH"

#
# Clean environment
module purge

#
# Load needed environment
module load git
module load cmake
module load gnu/8.4.0
module load zlib/1.2.11--gnu--8.4.0
module load szip/2.1.1--gnu--8.4.0
module load hdf5/1.12.0--gnu--8.4.0
module load spectrum_mpi/10.4.0--binary
module unload hdf5/1.12.0--gnu--8.4.0
module load hdf5/1.12.0--gnu--8.4.0

#
# Run all the test cases
for nRows in $(ls | grep $suffix) ; do
	cd $nRows
	echo "processing "${nRows} >> ${2}/run_${1}-${4}.log
	nrow=${nRows%"$suffix"}
	nMaxBlocks=$(echo $nrow $minNrows | awk '{printf "%d\n", $1*1000 / $2}')
	algs=(${algorithms[@]})
	for nBlocks in ${BLKblocks[@]}; do
		if [[ $nMaxBlocks -ge $nBlocks ]]; then
			algs+=("BlockRowPerm_"$nBlocks)
		fi
	done
	echo "applied algoithms: "${algs[@]} >> ${2}/run_${1}-${4}.log
	
	for i in *.h5 ; do
		prefix=`echo ${i} | sed -e 's/\_matVec.*$//'`
		idx=`echo ${i} | sed -e '{s/^.*\_matVec_\([0-9]*\).h5$/\1/};{s/^[0]*//}'`
		if ! [[ -v $idx ]]; then
			idx=0
		fi
		echo ${i} ${prefix} ${idx} ${PWD} >> ${2}/run_${1}-${4}.log

		for alg in ${algs[@]} ; do  
			date >> ${2}/run_${1}-${4}.log
			echo ${alg} >> ${2}/run_${1}-${4}.log
			if [[ ${4} == *"instr"* ]]; then
				echo "mpirun --bind-to core -n "$SLURM_JOB_CPUS_PER_NODE" "${exe}" -d "." -p "${prefix}" -n "matVec"  -i "${idx}" -k "${alg}" -o timings_"${alg}"_"${4}".sqlite.db" >> ${2}/run_${1}-${4}.log
				echo "---------------------------------------------------------------------------------------------------"
				mpirun --bind-to core -n $SLURM_JOB_CPUS_PER_NODE ${exe} -d "." -p "${prefix}" -n "matVec"  -k ${alg} -i ${idx} -o timings_${alg}_${4}.sqlite.db
			else
				echo "mpirun --bind-to core -n "$SLURM_JOB_CPUS_PER_NODE" "${exe}" -d "." -p "${prefix}" -n "matVec"  -i "${idx}" -k "${alg}" -o timings_"${alg}".sqlite.db" >> ${2}/run_${1}-${4}.log
				echo "---------------------------------------------------------------------------------------------------"
				mpirun --bind-to core -n $SLURM_JOB_CPUS_PER_NODE ${exe} -d "." -p "${prefix}" -n "matVec"  -k ${alg} -i ${idx} -o timings_${alg}.sqlite.db
			fi
		done
	done
	cd ${2}
done


date >> ${2}/run_${1}-${4}.log

#
# Exit with no error
exit 0;
